from mmhuman3d.core.conventions.keypoints_mapping.human_data import (  # noqa:F401,E501
    APPROXIMATE_MAP, APPROXIMATE_MAPPING_LIST, HUMAN_DATA, HUMAN_DATA_BODY,
    HUMAN_DATA_HEAD, HUMAN_DATA_HIP, HUMAN_DATA_LEFT_HAND, HUMAN_DATA_LIMBS,
    HUMAN_DATA_LIMBS_INDEX, HUMAN_DATA_PALETTE, HUMAN_DATA_PARTS,
    HUMAN_DATA_RIGHT_HAND, HUMAN_DATA_SHOULDER,
)

HUMAN_DATA_LIMB_NAMES = {
    'left_ankle': {
        'left_knee': 'left_lower_leg'
    },
    'right_ankle': {
        'right_knee': 'right_lower_leg'
    },
    'left_shoulder': {
        'left_elbow': 'left_upperarm'
    },
    'right_shoulder': {
        'right_elbow': 'right_upperarm'
    },
    'left_elbow': {
        'left_wrist': 'left_forearm'
    },
    'right_elbow': {
        'right_wrist': 'right_forearm'
    },
    'left_hip_extra': {
        'left_knee': 'left_thigh'
    },
    'right_hip_extra': {
        'right_knee': 'right_thigh'
    },
}

HUMAN_DATA_FOOT = [
    'left_ankle', 'left_foot', 'left_heel', 'left_ankle_openpose',
    'left_bigtoe_openpose', 'left_smalltoe_openpose', 'left_toe_3dhp',
    'left_bigtoe', 'left_smalltoe', 'right_ankle', 'right_foot', 'right_heel',
    'right_ankle_openpose', 'right_bigtoe_openpose', 'right_smalltoe_openpose',
    'right_toe_3dhp', 'right_bigtoe', 'right_smalltoe'
]

HUMAN_DATA_PARTS['foot'] = HUMAN_DATA_FOOT



# yapf: disable
import logging
import numpy as np
import os
import time
import torch
from prettytable import PrettyTable
from torch.utils.data import DataLoader
from typing import List, Tuple, Union
from xrprimer.utils.log_utils import get_logger

from xrmocap.core.evaluation.align_keypoints3d import align_keypoints3d
from xrmocap.data_structure.keypoints import Keypoints
from xrmocap.model.architecture.base_architecture import BaseArchitecture
from xrmocap.utils.distribute_utils import collect_results, is_main_process
from xrmocap.utils.mvp_utils import (
    AverageMeter, convert_result_to_kps, norm2absolute,
)
from .metrics import calc_limbs_accuracy, evaluate

# yapf: enable


class MVPEvaluation:
    """Evaluation for MvP method."""

    def __init__(
        self,
        test_loader: DataLoader,
        selected_limbs_name: List[List[str]] = [],
        additional_limbs_names: List[List[str]] = [],
        n_max_person: int = 10,
        dataset_name: Union[None, str] = None,
        pred_kps3d_convention: str = 'coco',
        gt_kps3d_convention: str = 'coco',
        print_freq: int = 100,
        final_output_dir: Union[None, str] = None,
        logger: Union[None, str, logging.Logger] = None,
    ):
        """Initialization for the class.

        Args:
            test_loader (DataLoader):
                Test dataloader.
            selected_limbs_name (List[List[str]], optional):
                The name of selected
                limbs in evaluation. Defaults to [].
            additional_limbs_names (List[List[str]], optional):
                Names at both ends of the limb. Defaults to [].
            n_max_person (int, optional):
                Number of maximum person the model can predict. Defaults to 10.
            dataset_name (Union[None, str], optional):
                Name of the dataset. Defaults to None.
            pred_kps3d_convention (str, optional):
                Keypoints3d convention of the predicted keypoints.
                Defaults to 'coco'.
            gt_kps3d_convention (str, optional):
                Keypoints3d convention of the ground-truth keypoints.
                Defaults to 'coco'.
            print_freq (int, optional):
                Printing frequency. Defaults to 100.
            final_output_dir (Union[None, str], optional):
                Directory to output folder. Defaults to None.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """

        self.logger = get_logger(logger)

        self.test_loader = test_loader
        self.dataset_name = dataset_name
        self.print_freq = print_freq
        self.output_dir = final_output_dir
        self.dataset = test_loader.dataset
        self.gt_n_frame = test_loader.dataset.len
        self.pred_kps3d_convention = pred_kps3d_convention
        self.n_max_person = n_max_person
        self.selected_limbs_name = selected_limbs_name
        self.additional_limbs_names = additional_limbs_names
        self.gt_kps3d_convention = gt_kps3d_convention

    def run(
        self,
        model: BaseArchitecture,
        threshold: float = 0.1,
        is_train: bool = False,
    ):

        # validate model and get predictions
        preds_single = self.model_validate(
            model,
            threshold=threshold,
            is_train=is_train,
        )
        kps3d_pred_list = collect_results(preds_single, len(self.dataset))

        # save predicted keypoints3d
        precision = None
        if is_main_process():
            n_frame = len(kps3d_pred_list)
            n_kps = kps3d_pred_list[0].shape[1]
            kps3d_pred = np.full((n_frame, self.n_max_person, n_kps, 4),
                                 np.nan)

            for frame_idx, per_frame_kps3d in enumerate(kps3d_pred_list):
                if len(per_frame_kps3d) > 0:
                    n_valid_person, keypoints3d_pred_valid = \
                        convert_result_to_kps([per_frame_kps3d])
                    kps3d_pred[
                        frame_idx, :n_valid_person] = keypoints3d_pred_valid

            keypoints3d_pred = Keypoints(
                dtype='numpy',
                kps=kps3d_pred,
                mask=kps3d_pred[..., -1] > 0,
                convention=self.pred_kps3d_convention,
                logger=self.logger)

            # convert gt to keypoints3d
            gt_n_person = np.max(
                np.array([
                    scene_keypoints.get_person_number()
                    for scene_keypoints in self.dataset.gt3d
                ]))
            gt_n_kps = self.dataset.gt3d[0].get_keypoints_number()
            kps3d_gt = np.full((self.gt_n_frame, gt_n_person, gt_n_kps, 4),
                               np.nan)
            mask_gt = np.full((self.gt_n_frame, gt_n_person, gt_n_kps), np.nan)

            start_frame = 0
            for keypoints3d_gt_scene in self.dataset.gt3d:

                kps3d_gt_scene = keypoints3d_gt_scene.get_keypoints().numpy(
                )  # [n_frame, n_person, n_kps, 4]
                mask_gt_scene = keypoints3d_gt_scene.get_mask().numpy(
                )  # [n_frame, n_person, n_kps]

                n_frame, n_person, _, _ = kps3d_gt_scene.shape
                end_frame = min(start_frame + n_frame, self.gt_n_frame)

                kps3d_gt[start_frame:end_frame, :n_person, ...] = \
                    kps3d_gt_scene[:end_frame-start_frame, ...]
                mask_gt[start_frame:end_frame, :n_person, ...] = \
                    mask_gt_scene[:end_frame-start_frame, ...]
                start_frame = end_frame

            keypoints3d_gt = Keypoints(
                dtype='numpy',
                kps=kps3d_gt,
                mask=mask_gt,
                convention=self.gt_kps3d_convention,
                logger=self.logger)

            if not is_train:
                pred_file = os.path.join(self.output_dir, 'pred_kps3d.npz')
                gt_file = os.path.join(self.output_dir, 'gt_kps3d.npz')

                self.logger.info(f'Saving 3D keypoints to: {pred_file}')
                keypoints3d_pred.dump(pred_file)
                self.logger.info(f'Saving 3D keypoints to: {gt_file}')
                keypoints3d_gt.dump(gt_file)

            # quantitative evaluation and print result
            if 'panoptic' in self.dataset_name:
                tb = PrettyTable()
                mpjpe_threshold = np.arange(25, 155, 25)

                aps, recs, mpjpe, recall500 = \
                    self.evaluate_map(keypoints3d_pred, keypoints3d_gt)

                tb.field_names = ['Threshold/mm'] + \
                    [f'{i}' for i in mpjpe_threshold]
                tb.add_row(['AP'] + [f'{ap * 100:.2f}' for ap in aps])
                tb.add_row(['Recall'] + [f'{re * 100:.2f}' for re in recs])
                tb.add_row(['recall@500mm'] +
                           [f'{recall500 * 100:.2f}' for re in recs])
                self.logger.info('\n' + tb.get_string())
                self.logger.info(f'MPJPE: {mpjpe:.2f}mm')

                precision = np.mean(aps[0])

            elif 'campus' in self.dataset_name \
                    or 'shelf' in self.dataset_name:

                pred_keypoints3d_, gt_keypoints3d_, limbs = align_keypoints3d(
                    keypoints3d_pred, keypoints3d_gt,
                    self.pred_kps3d_convention, self.selected_limbs_name,
                    self.additional_limbs_names)
                check_result, eval_table = calc_limbs_accuracy(
                    pred_keypoints3d_,
                    gt_keypoints3d_,
                    limbs,
                    logger=self.logger)

                person_wise_avg = np.sum(
                    check_result > 0, axis=(0, 2)) / np.sum(
                        np.abs(check_result), axis=(0, 2))
                precision = np.sum(person_wise_avg) / len(person_wise_avg)

                self.logger.info('\n' + eval_table.get_string())
                evel_dict = evaluate(
                    pred_keypoints3d_,
                    gt_keypoints3d_,
                    scale=1.,
                    logger=self.logger)
                self.logger.info('MPJPE: {:.2f} ± {:.2f} mm'.format(
                    evel_dict['mpjpe_mean'], evel_dict['mpjpe_std']))
                self.logger.info(
                    f'PA-MPJPE: {evel_dict["pa_mpjpe_mean"]:.2f} ± '
                    f'{evel_dict["pa_mpjpe_std"]:.2f} mm')
                self.logger.info(f'PCK@50mm: {evel_dict["pck"][50]:.2f} %')
                self.logger.info(f'PCK@100mm: {evel_dict["pck"][100]:.2f} %')

            else:
                self.logger.warning(f'Dataset {self.dataset_name} '
                                    'is not yet implemented.')
                raise NotImplementedError

        return precision

    def model_validate(self,
                       model: BaseArchitecture,
                       threshold: float,
                       is_train: bool = False):
        """Validate model during training or testing.

        Args:
            model (BaseArchitecture):
                Model to be evaluated.
            threshold (float):
                Confidence threshold to filter non-human keypoints.
            is_train (bool, optional):
                True if it is called during trainig. Defaults to False.

        Returns:
            float: model evaluation result, precision
        """
        batch_time = AverageMeter()
        data_time = AverageMeter()

        model.eval()

        preds = []
        with torch.no_grad():
            end = time.time()
            for i, (inputs, meta) in enumerate(self.test_loader):
                data_time.update(time.time() - end)
                assert len(inputs) == self.dataset.n_views
                output = model(views=inputs, meta=meta)

                gt_kps3d = meta[0]['kps3d'].float()
                n_kps = gt_kps3d.shape[2]
                bs, n_queries = output['pred_logits'].shape[:2]

                src_poses = output['pred_poses']['outputs_coord']. \
                    view(bs, n_queries, n_kps, 3)
                src_poses = norm2absolute(src_poses, model.module.grid_size,
                                          model.module.grid_center)
                score = output['pred_logits'][:, :, 1:2].sigmoid()
                score = score.unsqueeze(2).expand(-1, -1, n_kps, -1)
                temp = (score > threshold).float() - 1

                pred = torch.cat([src_poses, temp, score], dim=-1)
                pred = pred.detach().cpu().numpy()
                for b in range(pred.shape[0]):
                    preds.append(pred[b])

                batch_time.update(time.time() - end)
                end = time.time()
                if (i % self.print_freq == 0 or i
                        == len(self.test_loader) - 1) and is_main_process():
                    gpu_memory_usage = torch.cuda.memory_allocated(0)
                    speed = len(inputs) * inputs[0].size(0) / batch_time.val
                    msg = f'Test: [{i}/{len(self.test_loader)}]\t' \
                        f'Time: {batch_time.val:.3f}s ' \
                        f'({batch_time.avg:.3f}s)\t' \
                        f'Speed: {speed:.1f} samples/s\t' \
                        f'Data: {data_time.val:.3f}s ' \
                        f'({data_time.avg:.3f}s)\t' \
                        f'Memory {gpu_memory_usage:.1f}'
                    self.logger.info(msg)

        return preds

    def evaluate_map(
        self,
        pred_keypoints3d: Keypoints,
        gt_keypoints3d: Keypoints,
        threshold: float = 0.1) \
            -> Tuple[List[float], List[float], float, float]:
        """Evaluate MPJPE, mAP and recall based on MPJPE. Mainly for panoptic
        predictions.

        Args:
            pred_keypoints3d (Keypoints):
                Predicted 3D keypoints.
            threshold (float):
                Threshold for valid keypoints. Defaults to 0.1.

        Returns:
            Tuple[List[float], List[float], float, float]:
                List of AP, list of recall, MPJPE value and recall@500mm.
        """

        pred_kps3d = pred_keypoints3d.get_keypoints()
        gt_kps3d = gt_keypoints3d.get_keypoints()
        gt_n_frame, _, _, _ = gt_kps3d.shape
        pred_n_frame, _, n_kps, _ = pred_kps3d.shape
        total_gt = 0
        eval_list = []
        if not pred_n_frame == gt_n_frame:
            self.logger.error(f'Frame number mismatch: {pred_n_frame} '
                              f'pred and {self.gt_n_frame} gt')
            raise ValueError

        trans_ground = np.array([[1.0, 0.0, 0.0], [0.0, 0.0, -1.0],
                                 [0.0, 1.0, 0.0]])

        for frame_idx_all_scenes in range(gt_n_frame):

            gt_frame_kps3d = gt_kps3d[frame_idx_all_scenes, :, :n_kps, :]
            check_valid = np.sum(gt_frame_kps3d, axis=1)  # [n_person, 4]
            gt_frame_kps3d = gt_frame_kps3d[check_valid[:, -1] > 0]

            gt_frame_kps3d = [
                gt_person_kps3d[:, 0:3].dot(trans_ground)
                for gt_person_kps3d in gt_frame_kps3d
            ]
            if len(gt_frame_kps3d) == 0:
                continue

            pred_frame_kps3d_valid = pred_kps3d[frame_idx_all_scenes].copy()

            for pred_person_kps3d in pred_frame_kps3d_valid:
                mpjpes = []
                for gt_person_kps3d in gt_frame_kps3d:
                    vis = gt_person_kps3d[:, -1] > threshold

                    mpjpe = np.mean(
                        np.sqrt(
                            np.sum(
                                (np.array(pred_person_kps3d[vis, 0:3]) -
                                 np.array(gt_person_kps3d[vis, 0:3]))**2,
                                axis=-1)))
                    mpjpes.append(mpjpe)
                min_gt = np.argmin(mpjpes)
                min_mpjpe = np.min(mpjpes)
                score = pred_person_kps3d[0, -1]
                eval_list.append({
                    'mpjpe': float(min_mpjpe),
                    'score': float(score),
                    'gt_id': int(total_gt + min_gt)
                })

            total_gt += len(gt_frame_kps3d)

        mpjpe_threshold = np.arange(25, 155, 25)
        aps = []
        recs = []
        for t in mpjpe_threshold:
            ap, rec = self._eval_list_to_ap(eval_list, total_gt, t)
            aps.append(ap)
            recs.append(rec)

        return \
            aps, \
            recs, \
            self._eval_list_to_mpjpe(eval_list), \
            self._eval_list_to_recall(eval_list, total_gt)

    def _eval_list_to_ap(self, eval_list, total_gt, threshold):
        """convert evaluation result to ap."""
        eval_list.sort(key=lambda k: k['score'], reverse=True)
        total_num = len(eval_list)

        tp = np.zeros(total_num)
        fp = np.zeros(total_num)
        gt_det = []
        for i, item in enumerate(eval_list):
            if item['mpjpe'] < threshold and item['gt_id'] not in gt_det:
                tp[i] = 1
                gt_det.append(item['gt_id'])
            else:
                fp[i] = 1
        tp = np.cumsum(tp)
        fp = np.cumsum(fp)
        recall = tp / (total_gt + 1e-5)
        precise = tp / (tp + fp + 1e-5)
        for n in range(total_num - 2, -1, -1):
            precise[n] = max(precise[n], precise[n + 1])

        precise = np.concatenate(([0], precise, [0]))
        recall = np.concatenate(([0], recall, [1]))
        index = np.where(recall[1:] != recall[:-1])[0]
        ap = np.sum((recall[index + 1] - recall[index]) * precise[index + 1])

        return ap, recall[-2]

    def _eval_list_to_mpjpe(self, eval_list, threshold=500):
        """convert evaluation result to mpjpe."""
        eval_list.sort(key=lambda k: k['score'], reverse=True)
        gt_det = []

        mpjpes = []
        for i, item in enumerate(eval_list):
            if item['mpjpe'] < threshold and item['gt_id'] not in gt_det:
                mpjpes.append(item['mpjpe'])
                gt_det.append(item['gt_id'])

        return np.mean(mpjpes) if len(mpjpes) > 0 else np.inf

    def _eval_list_to_recall(self, eval_list, total_gt, threshold=500):
        """convert evaluation result to recall."""
        gt_ids = [e['gt_id'] for e in eval_list if e['mpjpe'] < threshold]

        return len(np.unique(gt_ids)) / total_gt



# yapf: disable
import copy
import logging
import numpy as np
import torch
from mmcv.runner import load_checkpoint
from torchvision.transforms import Compose
from tqdm import tqdm
from typing import List, Tuple, Union, overload
from xrprimer.data_structure.camera import FisheyeCameraParameter
from xrprimer.utils.log_utils import get_logger

from xrmocap.data_structure.body_model import SMPLData
from xrmocap.data_structure.keypoints import Keypoints
from xrmocap.io.image import (
    get_n_frame_from_mview_src, load_clip_from_mview_src,
)
from xrmocap.model.architecture.builder import build_architecture
from xrmocap.model.registrant.builder import SMPLify, build_registrant
from xrmocap.transform.image.builder import build_image_transform
from xrmocap.transform.image.shape import get_affine_trans_aug
from xrmocap.transform.keypoints3d.optim.builder import (
    BaseOptimizer, build_keypoints3d_optimizer,
)
from xrmocap.utils.geometry import get_scale
from xrmocap.utils.mvp_utils import norm2absolute, process_dict
from .mperson_smpl_estimator import MultiPersonSMPLEstimator

# yapf: enable


class MultiViewMultiPersonEnd2EndEstimator(MultiPersonSMPLEstimator):
    """Api for estimating keypoints3d and smpl in a multi-view multi-person
    scene, using end2end learning-based method."""

    def __init__(
        self,
        work_dir: str,
        img_pipeline: dict,
        image_size: List[int],
        heatmap_size: List[int],
        n_max_person: int,
        n_kps: int,
        kps3d_convention: str,
        kps3d_model: Union[dict, torch.nn.Module],
        kps3d_model_path: Union[None, str],
        dataset: Union[None, str] = None,
        inference_conf_thr: float = 0.0,
        load_batch_size: int = 1,
        kps3d_optimizers: Union[List[Union[BaseOptimizer, dict]], None] = None,
        smplify: Union[None, dict, SMPLify] = None,
        cam_metric_unit: str = 'millimeter',
        verbose: bool = True,
        logger: Union[None, str, logging.Logger] = None,
    ) -> None:
        """Initialization of the class.

        Args:
            work_dir (str):
                Path to the folder for running the api. No file in work_dir
                will be modified
                or added by MultiViewSinglePersonSMPLEstimator.
            img_pipeline (dict):
                Image pre-processin pipeline for input images,
                need to align with configs during model training.
            dataset (str):
                Name of the dataset.
            image_size (List[int]):
                Input image size, need to align with configs during
                    model training.
            heatmap_size (List[int]):
                Heatmap size, need to align with configs during model training.
            n_max_person (int):
                Number of maximum person the model can handle.
            n_kps (int):
                Number of keypoints.
            kps3d_convention (str):
                Convention of keypoints.
            kps3d_model (Union[dict, torch.nn.Module]):
                An end-to-end mview mperson keypoints3d predicting model.
            kps3d_model_path (Union[None, str]):
                Path to pre-trained model for inference.
            inference_conf_thr (float, optional):
                Threshold for confidence whether prediction is a human.
                Defaults to 0.0.
            load_batch_size (int, optional):
                Batch size in image loading. Defaults to 1.
            kps3d_optimizers (Union[List[Union[BaseOptimizer, dict]],
                None], optional):
                A list of keypoints3d optimizers or their configs. If given,
                keypoints3d will be
                optimized by the cascaded final optimizers before estimation.
                Defaults to None.
            smplify (Union[None, dict, SMPLify], optional):
                A SMPLify instance or its config. Defaults to None.
            cam_metric_unit (str, optional):
                Metric unit for camera parameters. Defaults to 'millimeter'.
            verbose (bool, optional):
                Whether to print(logger.info) information during estimating.
                Defaults to True.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """

        super().__init__(
            work_dir=work_dir,
            smplify=smplify,
            kps3d_optimizers=kps3d_optimizers,
            verbose=verbose,
            logger=logger)

        if torch.cuda.is_available():
            self.logger.info('CUDA is available')
        else:
            self.logger.warning('CUDA is not available, running on CPU')

        self.logger = get_logger(logger)

        self.dataset = dataset
        self.image_size = np.array(image_size)
        self.heatmap_size = np.array(heatmap_size)
        self.inference_conf_thr = inference_conf_thr
        self.kps3d_convention = kps3d_convention
        self.n_max_person = n_max_person
        self.n_kps = n_kps

        if cam_metric_unit == 'meter':
            self.factor = 1.0
        elif cam_metric_unit == 'centimeter':
            self.factor = 100.0
        elif cam_metric_unit == 'millimeter':
            self.factor = 1000.0

        # mvp model only accept batch size = 1 and views in the format of list
        self.load_batch_size = load_batch_size
        if self.load_batch_size != 1:
            self.logger.error('Please set load_batch_size to' +
                              '1 for end2end estimator')
            raise ValueError

        if isinstance(kps3d_model, dict):
            if kps3d_model_path is None:
                self.logger.error('Please define a pretrained model')
                raise ValueError
            kps3d_model_setup = dict(
                type='MviewPoseTransformer',
                is_train=False,
                logger=self.logger)
            kps3d_model_setup.update(kps3d_model)
            self.kps3d_model = build_architecture(kps3d_model_setup)
            self.logger.info(f'Load saved models state {kps3d_model_path}')
            load_checkpoint(
                self.kps3d_model,
                kps3d_model_path,
                logger=self.logger,
                map_location='cpu')
        else:
            self.kps3d_model = kps3d_model

        self.img_pipeline = []
        for transform in img_pipeline:
            if isinstance(transform, dict):
                transform = build_image_transform(transform)
            self.img_pipeline.append(transform)
        self.img_pipeline = Compose(self.img_pipeline)

        if isinstance(smplify, dict):
            smplify['logger'] = logger
            self.smplify = build_registrant(smplify)
        else:
            self.smplify = smplify
        if kps3d_optimizers is None:
            self.kps3d_optimizers = None
        else:
            self.kps3d_optimizers = []
            for kps3d_optim in kps3d_optimizers:
                if isinstance(kps3d_optim, dict):
                    kps3d_optim['logger'] = logger
                    kps3d_optim = build_keypoints3d_optimizer(kps3d_optim)
                self.kps3d_optimizers.append(kps3d_optim)

    @overload
    def run(
        self, img_arr: np.ndarray, cam_param: List[FisheyeCameraParameter]
    ) -> Tuple[List[Keypoints], Keypoints, SMPLData]:
        ...

    @overload
    def run(
        self, img_paths: List[List[str]],
        cam_param: List[FisheyeCameraParameter]
    ) -> Tuple[List[Keypoints], Keypoints, SMPLData]:
        ...

    @overload
    def run(
        self, video_path: List[str], cam_param: List[FisheyeCameraParameter]
    ) -> Tuple[List[Keypoints], Keypoints, SMPLData]:
        ...

    def run(
        self,
        cam_param: List[FisheyeCameraParameter],
        img_arr: Union[None, np.ndarray] = None,
        img_paths: Union[None, List[List[str]]] = None,
        video_paths: Union[None, List[str]] = None,
    ) -> Tuple[List[Keypoints], List[SMPLData]]:
        """Run mutli-view multi-person end2end estimator once. run() needs one
        images input among [img_arr, img_paths, video_paths].

        Args:
            cam_param (List[FisheyeCameraParameter]):
                A list of FisheyeCameraParameter instances.
            img_arr (Union[None, np.ndarray], optional):
                A multi-view image array, in shape
                [n_view, n_frame, h, w, c]. Defaults to None.
            img_paths (Union[None, List[List[str]]], optional):
                A nested list of image paths, in shape
                [n_view, n_frame]. Defaults to None.
            video_paths (Union[None, List[str]], optional):
                A list of video paths, each is a view.
                Defaults to None.

        Returns:
            Tuple[List[Keypoints], List[SMPLData]]:
                A list of keypoints3d, a list of SMPLData.
        """
        input_list = [img_arr, img_paths, video_paths]
        input_count = 0

        for input_instance in input_list:
            if input_instance is not None:
                input_count += 1
        if input_count > 1:
            self.logger.error('Redundant input!\n' +
                              'Please offer only one among' +
                              ' img_arr, img_paths and video_paths.')
            raise ValueError
        elif input_count < 1:
            self.logger.error('No image input has been found!\n' +
                              'img_arr, img_paths and video_paths are None.')
            raise ValueError
        n_frame = get_n_frame_from_mview_src(img_arr, img_paths, video_paths,
                                             self.logger)

        kps3d_batch = np.full((n_frame, self.n_max_person, self.n_kps, 4),
                              np.nan)

        # load from scene input
        for start_idx in tqdm(range(0, n_frame, self.load_batch_size)):
            end_idx = min(n_frame, start_idx + self.load_batch_size)
            mview_batch_arr = load_clip_from_mview_src(
                start_idx=start_idx,
                end_idx=end_idx,
                img_arr=img_arr,
                img_paths=img_paths,
                video_paths=video_paths,
                logger=self.logger)
            _, _, h, w, c = mview_batch_arr.shape

            # prepare input data, img list and meta list
            list_inputs = []
            for _, img in enumerate(mview_batch_arr.squeeze()):
                img_tensor = self.img_pipeline(img)
                img_tensor = img_tensor.unsqueeze(0)
                list_inputs.append(img_tensor)
            meta = self.prepare_meta(cam_param, h, w)

            frame_valid_pred_kps3d, _ = self.estimate_perception3d(
                list_inputs, meta, self.inference_conf_thr)

            pred_n_person = frame_valid_pred_kps3d.shape[0]
            kps3d_batch[start_idx, :pred_n_person,
                        ...] = frame_valid_pred_kps3d

        # Convert array to keypoints instance
        pred_keypoints3d = Keypoints(
            dtype='numpy',
            kps=kps3d_batch,
            mask=kps3d_batch[..., -1] > 0,
            convention=self.kps3d_convention,
            logger=self.logger)

        # Optimizing keypoints3d
        pred_keypoints3d = self.optimize_keypoints3d(pred_keypoints3d)

        # Fitting SMPL model
        smpl_data_list = self.estimate_smpl(keypoints3d=pred_keypoints3d)

        return pred_keypoints3d, smpl_data_list

    def estimate_perception3d(self,
                              img_arr: Union[None, np.ndarray],
                              meta: Union[None, dict],
                              threshold: float = 0.0):
        """Estimate perception3d from images per frame.

        Args:
            img_arr (Union[None, np.ndarray]):
                Input images.
            meta (Union[None, dict]):
                Input meta data.
            threshold (float, optional):
                Threshold for a valid person. Defaults to 0.0.

        Returns:
            Union[np.ndarray, torch.Tensor]:
                All predicted 3D keypoints per frame [n_person, n_kps, 4]
        """
        if torch.cuda.is_available():
            self.kps3d_model.cuda()
            img_arr = [i.to('cuda') for i in img_arr]
            meta = [
                process_dict(t, device='cuda', dummy_dim=None, dtype=None)
                for t in meta
            ]
        self.kps3d_model.eval()

        output = self.kps3d_model(views=img_arr, meta=meta)

        bs, n_queries = output['pred_logits'].shape[:2]
        src_poses = output['pred_poses']['outputs_coord']. \
            view(bs, n_queries, self.n_kps, 3)
        src_poses = norm2absolute(src_poses, self.kps3d_model.grid_size,
                                  self.kps3d_model.grid_center)

        if 'panoptic' in self.dataset:
            trans_ground = torch.tensor([[1.0, 0.0, 0.0], [0.0, 0.0, 1.0],
                                         [0.0, -1.0, 0.0]]).double()
            src_poses = torch.mm(src_poses, trans_ground)

        score = output['pred_logits'][:, :, 1:2].sigmoid()
        score = score.unsqueeze(2).expand(-1, -1, self.n_kps, -1)

        pred_kps3d = torch.cat([src_poses, score - threshold], dim=-1)
        pred_kps3d = pred_kps3d.detach().cpu().numpy()

        frame_pred_kps3d = pred_kps3d[0]
        frame_valid_pred_kps3d = frame_pred_kps3d[frame_pred_kps3d[:, 0,
                                                                   3] >= 0]

        return frame_valid_pred_kps3d / self.factor, frame_pred_kps3d

    def prepare_meta(self, cam_param: List[FisheyeCameraParameter],
                     height: int, width: int):
        """Prepare meta data for end2end estimator.

        Args:
            cam_param (List[FisheyeCameraParameter]):
                Camera parameters.
            height (int):
                Image height.
            width (int):
                Image width.
        """
        meta = []
        for cam_idx, input_camera in enumerate(cam_param):

            camera = input_camera.clone()
            if not camera.world2cam:
                camera.inverse_extrinsic()

            kw_data = {}
            k_tensor = torch.tensor(camera.get_intrinsic(k_dim=3))
            r_tensor = torch.tensor(camera.get_extrinsic_r())
            t_tensor = self.factor * torch.tensor(camera.get_extrinsic_t())

            k_tensor = k_tensor.double()
            r_tensor = r_tensor.double()
            t_tensor = t_tensor.double()

            dist_coeff_tensor = torch.tensor(camera.get_dist_coeff())

            c = np.array([width / 2.0, height / 2.0])
            s = get_scale(np.array([width, height]), self.image_size)
            r = 0  # NOTE: do not apply rotation augmentation
            hm_scale = self.heatmap_size / self.image_size

            # Affine transformations
            trans, inv_trans, aug_trans = get_affine_trans_aug(
                c, s, hm_scale, self.image_size, r)
            kw_data['affine_trans'] = trans
            kw_data['inv_affine_trans'] = inv_trans
            kw_data['aug_trans'] = aug_trans
            kw_data['center'] = c
            kw_data['scale'] = s

            view_kw_data = copy.deepcopy(kw_data)

            # Camera parameters
            view_kw_data['camera'] = dict()
            if 'panoptic' in self.dataset:
                trans_ground = torch.tensor([[1.0, 0.0, 0.0], [0.0, 0.0, -1.0],
                                             [0.0, 1.0, 0.0]]).double()
            elif 'shelf' or 'campus' in self.dataset:
                trans_ground = torch.tensor([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0],
                                             [0.0, 0.0, 1.0]]).double()
            else:
                self.logger.warning(
                    'This dataset is not supported by default, '
                    'please set the ground coordinate transformation. '
                    'Using Eye matrix by default.')
                trans_ground = torch.tensor([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0],
                                             [0.0, 0.0, 1.0]]).double()

            r_trans = torch.mm(r_tensor, trans_ground)
            t_trans = -torch.mm(r_trans.T, t_tensor.reshape((3, 1)))

            view_kw_data['camera']['camera_standard_T'] = t_tensor  # world2cam
            view_kw_data['camera']['R'] = r_trans
            view_kw_data['camera']['T'] = t_trans  # cam2world
            view_kw_data['camera']['K'] = k_tensor
            view_kw_data['camera']['dist_coeff'] = dist_coeff_tensor

            meta.append(view_kw_data)

        meta = [
            process_dict(t, device=None, dummy_dim=0, dtype='tensor')
            for t in meta
        ]

        return meta



# yapf: disable
import torch
from typing import Union

from xrmocap.transform.convention.keypoints_convention import (
    get_keypoint_idxs_by_part,
)
from .smplify import SMPLify

# yapf: enable


class SMPLifyX(SMPLify):
    """Re-implementation of SMPLify-X with extended features."""
    OPTIM_PARAM = SMPLify.OPTIM_PARAM + [
        'left_hand_pose', 'right_hand_pose', 'expression', 'jaw_pose',
        'leye_pose', 'reye_pose'
    ]

    def __set_keypoint_indexes__(self) -> None:
        """Set keypoint indexes to 1) body parts to be assigned different
        weights 2) be ignored for keypoint loss computation.

        Returns:
            None
        """
        SMPLify.__set_keypoint_indexes__(self)
        convention = self.body_model.keypoint_convention
        # head keypoints include all facial landmarks
        self.face_keypoint_idxs = get_keypoint_idxs_by_part(
            'head', convention=convention)

        left_hand_keypoint_idxs = get_keypoint_idxs_by_part(
            'left_hand', convention=convention)
        right_hand_keypoint_idxs = get_keypoint_idxs_by_part(
            'right_hand', convention=convention)
        self.hand_keypoint_idxs = [
            *left_hand_keypoint_idxs, *right_hand_keypoint_idxs
        ]

        self.body_keypoint_idxs = get_keypoint_idxs_by_part(
            'body', convention=convention)

        self.shoulder_keypoint_idxs = get_keypoint_idxs_by_part(
            'shoulder', convention=convention)

        self.hip_keypoint_idxs = get_keypoint_idxs_by_part(
            'hip', convention=convention)

        self.foot_keypoint_idxs = get_keypoint_idxs_by_part(
            'foot', convention=convention)

    def get_keypoint_weight(self,
                            use_shoulder_hip_only: bool = False,
                            body_weight: float = 1.0,
                            hand_weight: float = 1.0,
                            face_weight: float = 1.0,
                            shoulder_weight: Union[float, None] = None,
                            hip_weight: Union[float, None] = None,
                            foot_weight: Union[float, None] = None,
                            **kwargs) -> torch.Tensor:
        """Get per keypoint weight.

        Args:
            use_shoulder_hip_only (bool, optional):
                Whether to use only shoulder and hip
                keypoints for loss computation. This is useful in the
                warming-up stage to find a reasonably good initialization.
                Defaults to False.
            body_weight (float, optional):
                Weight of body keypoints. Body part segmentation
                definition is included in the HumanData convention.
                Defaults to 1.0.
            hand_weight (float, optional):
                Weight of hands keypoints. Body part segmentation
                definition is included in the HumanData convention.
                Defaults to 1.0.
            face_weight (float, optional):
                Weight of face keypoints. Body part segmentation
                definition is included in the HumanData convention.
                Defaults to 1.0.
            shoulder_weight (float, optional):
                Weight of shoulder keypoints. Body part segmentation
                definition is included in the HumanData convention.
                Defaults to None.
            hip_weight (float, optional):
                Weight of hip keypoints. Body part segmentation
                definition is included in the HumanData convention.
                Defaults to None.
            foot_weight (float, optional):
                Weight of feet keypoints. Body part segmentation
                definition is included in the HumanData convention.
                Defaults to None.

        Returns:
            torch.Tensor: Per keypoint weight tensor of shape (K).
        """
        n_keypoints = self.body_model.get_joint_number()

        # 3rd priority: set body parts weight manually
        # when both body weight and body parts weight set,
        # body parts weight override the body weight

        weight = torch.ones([n_keypoints]).to(self.device)

        # "body": includes "shoulder", "hip" and "foot"
        weight[self.body_keypoint_idxs] = \
            weight[self.body_keypoint_idxs] * body_weight

        if shoulder_weight is not None:
            weight[self.shoulder_keypoint_idxs] = 1.0
            weight[self.shoulder_keypoint_idxs] = \
                weight[self.shoulder_keypoint_idxs] * shoulder_weight

        if hip_weight is not None:
            weight[self.hip_keypoint_idxs] = 1.0
            weight[self.hip_keypoint_idxs] = \
                weight[self.hip_keypoint_idxs] * hip_weight

        if foot_weight is not None:
            weight[self.foot_keypoint_idxs] = 1.0
            weight[self.foot_keypoint_idxs] = \
                weight[self.foot_keypoint_idxs] * foot_weight

        weight[self.hand_keypoint_idxs] = \
            weight[self.hand_keypoint_idxs] * hand_weight
        weight[self.face_keypoint_idxs] = \
            weight[self.face_keypoint_idxs] * face_weight

        # 2nd priority: use_shoulder_hip_only
        if use_shoulder_hip_only:
            weight = torch.zeros([n_keypoints]).to(self.device)
            weight[self.shoulder_hip_keypoint_idxs] = 1.0
            if shoulder_weight is not None and hip_weight is not None and \
                    body_weight * face_weight * hand_weight == 0.0:
                weight[self.shoulder_keypoint_idxs] = \
                    weight[self.shoulder_keypoint_idxs] * shoulder_weight
                weight[self.hip_keypoint_idxs] = \
                    weight[self.hip_keypoint_idxs] * hip_weight
            else:
                self.logger.error(
                    'use_shoulder_hip_only is deprecated, '
                    'please manually set: body_weight=0.0, face_weight=0.0, '
                    'hand_weight=0.0, shoulder_weight=1.0, hip_weight=1.0 to '
                    'make sure correct weights are set.')
                raise ValueError

        # 1st priority: keypoints ignored
        if hasattr(self, 'ignore_keypoint_idxs'):
            weight[self.ignore_keypoint_idxs] = 0.0

        return weight



import cv2
import logging
import mediapipe as mp
import numpy as np
from tqdm import tqdm
from typing import List, Tuple, Union
from xrprimer.utils.ffmpeg_utils import video_to_array
from xrprimer.utils.log_utils import get_logger

from xrmocap.data_structure.keypoints import Keypoints
from xrmocap.transform.convention.keypoints_convention import get_keypoint_num


class MediapipeEstimator():

    def __init__(self,
                 mediapipe_kwargs: dict,
                 bbox_thr: float = 0.0,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Init a detector from mediapipe.

        Args:
            mediapipe_kwargs (dict):
                A dict contains args of mediapipe.
                refer to https://google.github.io/mediapipe/solutions/pose.html
                in detail.
            bbox_thr (float, optional):
                Threshold of a bbox. Those have lower scores will be ignored.
                Defaults to 0.0.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        # build the pose model
        mp_pose = mp.solutions.pose
        self.pose_model = mp_pose.Pose(**mediapipe_kwargs)
        self.bbox_thr = bbox_thr
        self.logger = get_logger(logger)
        self.convention = 'mediapipe_body'

    def get_keypoints_convention_name(self) -> str:
        """Get data_source from dataset type of the pose model.

        Returns:
            str:
                Name of the keypoints convention. Must be
                a key of KEYPOINTS_FACTORY.
        """
        return self.convention

    def infer_single_img(self, img_arr: np.ndarray, bbox_list: list):
        """Infer a single img with bbox.

        Args:
            image_array (Union[np.ndarray, list]):
                BGR image ndarray in shape [height, width, 3],
            bbox_list (Union[tuple, list]):
                A list of human bboxes.
                Shape of the nested lists is (n_human, 5).
                Each bbox is a bbox_xyxy with a bbox_score at last.

        Returns:
            multi_kps2d (list):
                A list of dict for human keypoints and bbox.
        """
        rt_bbox_list = []
        rt_kps2d_list = []
        for bbox in bbox_list:
            kps2d = None
            if bbox[4] > self.bbox_thr:
                img = img_arr[int(bbox[1]):int(bbox[3]),
                              int(bbox[0]):int(bbox[2])]
                result_mp = self.pose_model.process(
                    cv2.cvtColor(img, cv2.COLOR_BGR2RGB))
                if result_mp.pose_landmarks:
                    kps_list = [[
                        landmark.x * img.shape[1] + bbox[0],
                        landmark.y * img.shape[0] + bbox[1],
                        landmark.visibility
                    ] for landmark in result_mp.pose_landmarks.landmark]
                    kps2d = np.array(kps_list)
            if kps2d is not None:
                rt_bbox_list.append(bbox)
                rt_kps2d_list.append(kps2d)
        return rt_bbox_list, rt_kps2d_list

    def infer_array(self,
                    image_array: Union[np.ndarray, list],
                    bbox_list: Union[tuple, list],
                    disable_tqdm: bool = False,
                    **kwargs) -> Tuple[list, list]:
        """Infer frames already in memory(ndarray type).

        Args:
            image_array (Union[np.ndarray, list]):
                BGR image ndarray in shape [n_frame, height, width, 3],
                or a list of image ndarrays in shape [height, width, 3] while
                len(list) == n_frame.
            bbox_list (Union[tuple, list]):
                A list of human bboxes.
                Shape of the nested lists is (n_frame, n_human, 5).
                Each bbox is a bbox_xyxy with a bbox_score at last.
            disable_tqdm (bool, optional):
                Whether to disable the entire progressbar wrapper.
                Defaults to False.

        Returns:
            Tuple[list, list]:
                keypoints_list (list):
                    A list of human keypoints.
                    Shape of the nested lists is
                    (n_frame, n_human, n_keypoints, 3).
                    Each keypoint is an array of (x, y, confidence).
                bbox_list (list):
                    A list of human bboxes.
                    Shape of the nested lists is (n_frame, n_human, 5).
                    Each bbox is a bbox_xyxy with a bbox_score at last.
                    It could be smaller than the input bbox_list,
                    if there's no keypoints detected in some bbox.
        """
        ret_kps_list = []
        ret_bbox_list = []
        n_frame = len(image_array)
        n_kps = get_keypoint_num(self.get_keypoints_convention_name())
        for frame_index in tqdm(range(0, n_frame), disable=disable_tqdm):
            img_arr = image_array[frame_index]
            bboxes_in_frame = []
            for idx, bbox in enumerate(bbox_list[frame_index]):
                if bbox[4] > 0.0:
                    bboxes_in_frame.append(bbox)
            if len(bboxes_in_frame) > 0:
                bbox_results, kps2d_results = self.infer_single_img(
                    img_arr, bboxes_in_frame)
                frame_kps_results = np.zeros(
                    shape=(
                        len(kps2d_results),
                        n_kps,
                        3,
                    ))
                frame_bbox_results = np.zeros(shape=(len(bbox_results), 5))
                for idx, (bbox, keypoints) in enumerate(
                        zip(bbox_results, kps2d_results)):
                    frame_bbox_results[idx] = bbox
                    frame_kps_results[idx] = keypoints
                frame_kps_results = frame_kps_results.tolist()
                frame_bbox_results = frame_bbox_results.tolist()
            else:
                frame_kps_results = []
                frame_bbox_results = []
            ret_kps_list += [frame_kps_results]
            ret_bbox_list += [frame_bbox_results]
        return ret_kps_list, None, ret_bbox_list

    def infer_frames(self,
                     frame_path_list: list,
                     bbox_list: Union[tuple, list],
                     disable_tqdm: bool = False,
                     load_batch_size: Union[None, int] = None,
                     **kwargs) -> Tuple[list, list]:
        """Infer frames from file.

        Args:
            frame_path_list (list):
                A list of frames' absolute paths.
            bbox_list (Union[tuple, list]):
                A list of human bboxes.
                Shape of the nested lists is (n_frame, n_human, 5).
                Each bbox is a bbox_xyxy with a bbox_score at last.
            disable_tqdm (bool, optional):
                Whether to disable the entire progressbar wrapper.
                Defaults to False.
            load_batch_size (Union[None, int], optional):
                How many frames are loaded at the same time.
                Defaults to None, load all frames in frame_path_list.

        Returns:
            Tuple[list, list]:
                keypoints_list (list):
                    A list of human keypoints.
                    Shape of the nested lists is
                    (n_frame, n_human, n_keypoints, 3).
                    Each keypoint is an array of (x, y, confidence).
                bbox_list (list):
                    A list of human bboxes.
                    Shape of the nested lists is (n_frame, n_human, 5).
                    Each bbox is a bbox_xyxy with a bbox_score at last.
                    It could be smaller than the input bbox_list,
                    if there's no keypoints detected in some bbox.
        """
        ret_kps_list = []
        ret_boox_list = []
        if load_batch_size is None:
            load_batch_size = len(frame_path_list)
        for start_idx in range(0, len(frame_path_list), load_batch_size):
            end_idx = min(len(frame_path_list), start_idx + load_batch_size)
            if load_batch_size < len(frame_path_list):
                self.logger.info(
                    'Processing mediapipe on frames' +
                    f'({start_idx}-{end_idx})/{len(frame_path_list)}')
            image_array_list = []
            for frame_abs_path in frame_path_list[start_idx:end_idx]:
                img_np = cv2.imread(frame_abs_path)
                image_array_list.append(img_np)
            batch_pose_list, _, batch_boox_list = \
                self.infer_array(
                    image_array=image_array_list,
                    bbox_list=bbox_list[start_idx:end_idx],
                    disable_tqdm=disable_tqdm)
            ret_kps_list += batch_pose_list
            ret_boox_list += batch_boox_list
        return ret_kps_list, None, ret_boox_list

    def infer_video(self,
                    video_path: str,
                    bbox_list: Union[tuple, list],
                    disable_tqdm: bool = False,
                    **kwargs) -> Tuple[list, list]:
        """Infer frames from a video file.

        Args:
            video_path (str):
                Path to the video to be detected.
            bbox_list (Union[tuple, list]):
                A list of human bboxes.
                Shape of the nested lists is (n_frame, n_human, 5).
                Each bbox is a bbox_xyxy with a bbox_score at last.
            disable_tqdm (bool, optional):
                Whether to disable the entire progressbar wrapper.
                Defaults to False.

        Returns:
            Tuple[list, list]:
                keypoints_list (list):
                    A list of human keypoints.
                    Shape of the nested lists is
                    (n_frame, n_human, n_keypoints, 3).
                    Each keypoint is an array of (x, y, confidence).
                bbox_list (list):
                    A list of human bboxes.
                    Shape of the nested lists is (n_frame, n_human, 5).
                    Each bbox is a bbox_xyxy with a bbox_score at last.
                    It could be smaller than the input bbox_list,
                    if there's no keypoints detected in some bbox.
        """
        image_array = video_to_array(input_path=video_path, logger=self.logger)
        ret_kps_list, _, ret_boox_list = self.infer_array(
            image_array=image_array,
            bbox_list=bbox_list,
            disable_tqdm=disable_tqdm)
        return ret_kps_list, None, ret_boox_list

    def get_keypoints_from_result(
            self, kps2d_list: List[list]) -> Union[Keypoints, None]:
        """Convert returned keypoints2d into an instance of class Keypoints.

        Args:
            kps2d_list (List[list]):
                A list of human keypoints, returned by
                infer methods.
                Shape of the nested lists is
                (n_frame, n_human, n_keypoints, 3).

        Returns:
            Union[Keypoints, None]:
                An instance of Keypoints with mask and
                convention, data type is numpy.
                If no one has been detected in any frame,
                a None will be returned.
        """
        # shape: (n_frame, n_human, n_keypoints, 3)
        n_frame = len(kps2d_list)
        human_count_list = [len(human_list) for human_list in kps2d_list]
        if len(human_count_list) > 0:
            n_human = max(human_count_list)
        else:
            n_human = 0
        n_keypoints = get_keypoint_num(self.get_keypoints_convention_name())
        if n_human > 0:
            kps2d_arr = np.zeros(shape=(n_frame, n_human, n_keypoints, 3))
            mask_arr = np.ones_like(kps2d_arr[..., 0], dtype=np.uint8)
            for f_idx in range(n_frame):
                if len(kps2d_list[f_idx]) <= 0:
                    mask_arr[f_idx, ...] = 0
                    continue
                for h_idx in range(n_human):
                    if h_idx < len(kps2d_list[f_idx]):
                        mask_arr[f_idx, h_idx, ...] = np.sign(
                            np.array(kps2d_list[f_idx][h_idx])[:, -1])
                        kps2d_arr[f_idx,
                                  h_idx, :, :] = kps2d_list[f_idx][h_idx]
                    else:
                        mask_arr[f_idx, h_idx, ...] = 0
            keypoints2d = Keypoints(
                kps=kps2d_arr,
                mask=mask_arr,
                convention=self.get_keypoints_convention_name(),
                logger=self.logger)
        else:
            keypoints2d = None
        return keypoints2d



# yapf: disable
import logging
import numpy as np
import torch
from typing import Union

from xrmocap.data_structure.keypoints import Keypoints
from xrmocap.ops.top_down_association.identity_tracking.builder import (
    BaseTracking, build_identity_tracking,
)
from .base_optimizer import BaseOptimizer

# yapf: enable


class RemoveDuplicate(BaseOptimizer):
    """This optimization tool helps remove duplicate identities by the L2
    distance between 3D keypoints for each person and add tracked identities to
    optimized 3D keypoints."""

    def __init__(
        self,
        verbose: bool = True,
        threshold: float = 2.0,
        keep: str = 'by_index',
        identity_tracking: Union[None, dict, BaseTracking] = None,
        logger: Union[None, str, logging.Logger] = None,
    ):
        """Initialization for RemoveDuplicate optimizor.

        Args:
            verbose (bool, optional):
                Whether to log info. Defaults to True.
            threshold (float, optional):
                Threshold for duplicate 3D keypoints. Defaults to 2.0.
            keep (str, optional):
                Mode to keep 3D keypoints, 'by_index' keeps the lowest index,
                'by_conf' keeps the highest confidence. Defaults to 'by_index'.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """

        super().__init__(verbose=verbose, logger=logger)
        self.threshold = threshold
        self.keep = keep

        if self.keep not in {'by_index', 'by_conf'}:
            self.logger.error(f'"{self.keep}" is not a valid mode. '
                              'Please select the mode between '
                              '"by_index" and "by_conf".')
            raise ValueError

        if isinstance(identity_tracking, dict):
            identity_tracking['logger'] = logger
            self.identity_tracking = build_identity_tracking(identity_tracking)
        else:
            self.identity_tracking = identity_tracking

    def optimize_keypoints3d(self, keypoints3d: Keypoints,
                             **kwargs: dict) -> Keypoints:
        """Forward function for RemoveDuplicate optimizor.

        Args:
            keypoints3d (Keypoints):
                Keypoints3d to be optimized.

        Returns:
            Keypoints: Optimized keypoints3d.
        """
        # get kps3d array
        keypoints3d_optim = keypoints3d.clone()
        kps3d = keypoints3d_optim.get_keypoints()
        n_frame, n_max_person, n_kps, _ = kps3d.shape

        kps3d_optim = np.full((n_frame, n_max_person, n_kps, 4), np.nan)

        for frame_idx in range(n_frame):
            kps3d_frame = kps3d[frame_idx, ...]
            kps3d_frame = kps3d_frame[~np.isnan(kps3d_frame[:, 0, 0])]

            # calculate distance and remove duplicate
            dist = self.get_kps3d_dist(kps3d_frame)

            remove_idxs = []
            keep_idxs = []
            for person_idx, per_person_dist in enumerate(dist):
                if person_idx not in remove_idxs \
                        and person_idx not in keep_idxs:
                    to_remove = np.where(
                        per_person_dist[:] < self.threshold)[0]  # index

                    if self.keep == 'by_conf':
                        # keep the highest confidence
                        keep_idx = to_remove[np.argmax(kps3d_frame[to_remove,
                                                                   0, -1])]

                    elif self.keep == 'by_index':
                        # keep the lowest index
                        keep_idx = person_idx

                    to_keep = np.array([keep_idx])
                    to_remove = np.setdiff1d(to_remove, to_keep)

                    remove_idxs.extend(list(to_remove))
                    keep_idxs.extend(list(to_keep))

            # identity tracking
            if self.identity_tracking is not None:
                curr_kps3d = kps3d_frame[keep_idxs, ...]
                frame_identities = self.identity_tracking.query(curr_kps3d)

                # save to Kps3d
                kps3d_optim[frame_idx, frame_identities,
                            ...] = kps3d_frame[keep_idxs, ...]
            else:
                # save to Kps3d
                n_optim_person = len(keep_idxs)
                kps3d_optim[frame_idx, :n_optim_person,
                            ...] = kps3d_frame[keep_idxs, ...]

        keypoints3d_optim.set_keypoints(kps3d_optim)
        keypoints3d_optim.set_mask(kps3d_optim[..., -1] > 0)

        return keypoints3d_optim

    def get_kps3d_dist(self,
                       kps3d: Union[torch.Tensor, np.ndarray],
                       p: int = 2) -> np.ndarray:
        """Calculate the distance between each set of keypoints3d.

        Args:
            kps3d (Union[torch.Tensor, np.ndarray]):
                keypoints3d of the current frame, [n_person, n_kps, 4].
            p (int, optional):
                 p value for the p-norm distance to calculate
                 between each vector pair. Defaults to 2.

        Returns:
            np.ndarray:
                a distance metrix of the size:
                [n_valid_person, n_valid_person].
        """
        n_valid_person = kps3d.shape[0]

        person = torch.tensor(kps3d[:, 0:3].reshape(n_valid_person, -1))
        dist = torch.cdist(person, person, p)  # l2 dist by default

        return dist
